module Main(main) where

import Array
import Char
import IO
import List
import Maybe
import Monad
import System
import Time
import Random(randomIO)
import Directory

import Control.Concurrent
import Control.Exception
import Data.IORef
import System.IO.Unsafe
import System.Console.GetOpt
import Data.Unique
import qualified Posix

import RSA(readSymbolicLink)
import ConfigFile
import Curses
import ErrorLog
import ExampleConf
import Filter
import Gale
import GenUtil
import Version
import Puff
import Charset
import EIO
import Screen


data Env = Env {
    envAction :: IO (),
    envVerbose :: Bool,
    envConfig :: (Maybe String),
    envJustArgs :: Bool,
    envErrorLog :: Maybe String,
    envNoPufflog :: Bool
    }


----------------
-- Configuration
----------------

getGaleDomain :: IO String
getGaleDomain = do
    gps <- varLookupAll "GALE_DOMAIN"
    case gps of
	(x:_) -> return x
	[] -> fmap Posix.nodeName Posix.getSystemID 
    

getGaleProxy :: IO [String]
getGaleProxy = do
    gps <- varLookupAll "GALE_PROXY"
    case gps of 
	[] -> do
	    v <- getGaleDomain
	    return $ hostStrings v
	(x:_) -> return (words x)

getGaleId :: IO String
getGaleId  = do
    gps <- varLookupAll "GALE_ID"
    case gps of 
	(g:_) -> return g
	[] -> do
	    n <- Posix.getLoginName
	    d <- getGaleDomain
	    return $ n ++ "@" ++ d

getGaleAliases :: IO [(String,Category)]
getGaleAliases = do
    v <- galeFile "aliases/"
    fs <- handle (\_ -> return []) $ getDirectoryContents v
    let f fn = do
	fc <- first [readFile (v ++ fn), readSymbolicLink (v ++ fn)]
	return (fn, catParseNew fc)
    ks <- mapM (\x -> tryMost (f x)) fs
    return [x | (Right x) <- ks] 

getIdInstance = do
    pid <- Posix.getProcessID
    n <- liftM Posix.nodeName Posix.getSystemID
    return $ n ++ "/" ++ show pid
    
    
varsLookupAll :: [String] -> IO [String]
varsLookupAll ss  = fmap concat $ sequence (map (\x -> lookupEnv ("GINSU_" ++ x)) ss ++ map configLookupList ss ++ map lookupEnv ss)  where
    lookupEnv s = handleMost (const $ return []) $ fmap (:[]) (getEnv s)

varsLookupElse :: [String] -> IO String -> IO String
varsLookupElse ss action = varsLookupAll ss >>= \x -> case x of
    (x:_) -> return x
    _ -> action
    
varLookupAll :: String -> IO [String]
varLookupAll s = varsLookupAll [s]

varLookupElse s = varsLookupElse [s]


{-# NOINLINE screenLock #-}
screenLock = unsafePerformIO $ newMVar ()

withLock :: MVar () -> IO a -> IO a
withLock l action = withMVar l (const action)

galeFile :: String -> IO String
galeFile fn = do
    e <- getEnv "HOME"
    return (e ++ "/.gale/" ++ fn)

getEditor = varsLookupElse ["VISUAL", "EDITOR"] (return "vi")


getTmpFile = do
    (v::Int) <- randomIO
    u <- newUnique
    return ("/tmp/ginsu.puff." ++ show (abs v) ++ "." ++ show (hashUnique u))


insertKeys ic s = do
    mapM_ (\c -> writeChan ic (Left (KeyChar c))) s


bugMsg = unlines [
    "There has been an internal error",
    "please inform <john@ugcs.caltech.edu> or file a bug report at",
    "<http://bugs.ofb.net/cgi-bin/bugzilla/enter_bug.cgi?product=Ginsu>",
    "Include the following text in any error report:"
    ]
    
    
main = do
    (flags, acats) <- ginsuOpts
    withErrorMessage bugMsg $ do
    logname <- case envErrorLog flags of 
	Just x -> return x
	Nothing -> galeFile "ginsu.errorlog" 
    withErrorLog logname $ withStartEndEntrys "Ginsu" $ do
    mainThd <- myThreadId
    Posix.installHandler Posix.sigINT (Posix.Catch (killThread mainThd)) Nothing
    

    cs <- configLookup "CHARSET"
    charsetSetup cs
    setErrorLogPutStr Charset.csHPutStr

    Control.Exception.bracket_ initCurses endWin $ do
    cursset Invisible

    ic <- newChan
    b <- configLookupBool "DISABLE_SIGWINCH"
    case (b,cursesSigWinch) of 
	(False,Just s) -> Posix.installHandler s (Posix.Catch (writeChan ic (Left KeyResize))) Nothing >> return ()
	_ -> return ()

    gs <- fmap (concat . map words) $ varLookupAll "GALE_SUBSCRIBE"
    gps <- getGaleProxy
    gc <- newGaleContext gps (nub $ gs ++ acats)
    doRender widgetEmpty
    pl <- forkIO $ puffLoop ic gc
    gl <- forkIO $ getchLoop ic
    yo <- newIORef (0::Int)
    plog <- galeFile "ginsu.pufflog"
    ops <- if envNoPufflog flags then return [] else 
	doRender (widgetCenter $  widgetText "Reading pufflog...") >> readPuffs plog
    ps <- newMVar ((reverse $ zip [0..] (reverse ops)))
    cache_render_r <- newIORef ([],-1)
    cache_filter_r <- newIORef []
    mark_set_r <- newIORef (listArray (0,9) (replicate 10 []))
    mapM_ (\n -> checkMark n >>= \f -> modifyIORef mark_set_r f) [1..9]
    next_r <- newIORef (length ops + 5)
    selected_r <- newIORef 1
    filter_r <- newIORef []
    message_r <- newIORef helpText
    redraw_r <- newIORef (return () :: IO ())
    unless (envNoPufflog flags) $ (forkIO $ pufflogLoop ps 0) >> return ()

    s <- varLookupElse "ON_STARTUP" (return "")
    insertKeys ic s 
    
    mainLoop gc ic yo ps selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r redraw_r
    destroyGaleContext gc
    killThread pl
    killThread gl
    ps <- liftM (map snd) $ readMVar ps
    unless (envNoPufflog flags) $ doRender (widgetCenter $ widgetText "Writing pufflog...") >>  writePufflog ps

pufflogLoop ps_r n = do
    threadDelay 30000000 
    mv <- fmap (fmap fst . listToMaybe) $ readMVar ps_r
    case mv of 
	Just n' | n' /= n -> do
	    ps <- liftM (map snd) $ readMVar ps_r
	    writePufflog ps
	    pufflogLoop ps_r n'
	_ -> pufflogLoop ps_r n

writePufflog ps = do
    plog <- galeFile "ginsu.pufflog"
    plsize <- fmap read $ varLookupElse "PUFFLOG_SIZE" (return "100")
    writePuffs plog (take plsize ps)


checkMark n = do
    v <- configLookup ("MARK_" ++ show n)
    return $ maybe id id (v >>= \x -> parseFilter x >>= \y -> return (// [(n - 1,[y])]))


puffLoop ic gc = do
    p <- galeNextPuff gc
    writeChan ic $ Right p
    puffLoop ic gc
    
getchLoop ic = getCh >>= \x -> writeChan ic (Left x) >> getchLoop ic


apHead f (x:xs) = f x:xs
apHead _ [] = []


expandAliases :: [Category] -> IO [Category]
expandAliases cats = do
    as <- getGaleAliases 
    gd <- getGaleDomain
    let ec x@(c,"") = de $ head $ [(tc ++ drop (length f) c,td) | (f,(tc,td)) <- as, f `isPrefixOf` c] ++ [x]
	ec x = x
	de (c,"") = (c,gd)
	de x = x
    return $ map ec cats 

mainLoop gc ic yor psr selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r redraw_r = redraw >> getKey redraw >>= \x -> case x of
    (KeyChar '\DC2') -> reconnectGaleContext gc >> continue
    (KeyChar 'q') -> return ()
    (KeyChar 'Q') -> return ()
    KeyEnter -> scrollPuffs 1 >> rcontinue
    (KeyChar '\n') -> scrollPuffs 1 >> rcontinue
    (KeyChar '\r') -> scrollPuffs 1 >> rcontinue
    KeyBackspace -> scrollPuffs (-1) >> rcontinue
    (KeyChar '\b')-> scrollPuffs (-1) >> rcontinue
    (KeyChar '\ENQ') -> scrollPuffs 1  >> rcontinue
    (KeyChar '\EM') -> scrollPuffs (-1)  >> rcontinue
    (KeyChar '\EOT') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys `div` 2) >> rcontinue
    (KeyChar '\NAK') -> scrSize >>= \(ys,_) ->  scrollPuffs (- (ys `div` 2)) >> rcontinue
    (KeyChar '\ACK') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys) >> rcontinue
    (KeyChar '\STX') -> scrSize >>= \(ys,_) ->  scrollPuffs (-ys) >> rcontinue
    KeyNPage -> scrSize >>= \(ys,_) -> scrollPuffs (ys) >> rcontinue
    KeyPPage -> scrSize >>= \(ys,_) -> scrollPuffs (-ys) >> rcontinue
    (KeyChar '\f') -> attempt (touchWin stdScr >> endWin) >> rcontinue
    KeyResize -> attempt endWin >> continue
    (KeyChar 'E') -> do
	gc <- galeFile "ginsu.config"
	e <- getEditor
	withProgram $ System.system (e ++ " " ++ gc)
	loadStdConfiguration ".gale/ginsu.config"
	rcontinue
    (KeyChar 'g') -> writeIORef selected_r 1 >> select_perhaps select_next >> writeIORef yor 0 >> rcontinue 
    (KeyChar 'G') -> do
	ps <- readMVar psr
	case ps of 
	    ((n,_):_) -> writeIORef selected_r n >> select_perhaps select_prev >> scroll_end >> rcontinue 
	    _ -> continue
    (KeyChar 'j') -> select_next >> center_puff >> rcontinue
    (KeyChar 'k') -> select_prev >> center_puff >> rcontinue
    KeyDown -> select_next >> center_puff >> rcontinue
    KeyUp -> select_prev >> center_puff >> rcontinue
    (KeyChar 'u') -> modifyFilter  (drop 1) >> rcontinue
    (KeyChar 'U') -> modifyFilter (const []) >> rcontinue
    (KeyChar '!') -> modifyFilter (apHead f) >> rcontinue where
	f (FilterNot x) = x
	f x = FilterNot x
    (KeyChar 'x') -> modifyFilter f >> rcontinue where
	f (x:y:r) = y:x:r
	f x = x
    (KeyChar 't') -> filter_thread >> rcontinue
    (KeyChar 'T') -> filter_thread >> rcontinue
    (KeyChar 'm') -> getKey redraw >>= \x -> case x of
	    (KeyChar n) | n `elem` "123456789" -> do
		fs <- readIORef filter_r
		modifyIORef mark_set_r (// ([(ord n - ord '1',fs)])) >> continue
	    key -> keyError "Invalid mark" key >> rcontinue
    (KeyChar 'a') -> do
	p <- sel_puff
	case p of 
	    Nothing -> continue
	    Just p -> case getSigner p of
		Just a -> do
		    modifyFilter (FilterAuthor a:)
		    rcontinue
		Nothing -> setMessage "no signed puff to get author from" >> continue
    (KeyChar 'c') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filter: " ""
	case v of 
	    Just f -> maybe (return ()) (\x -> modifyFilter (x:)) $ parseFilter f 
	    Nothing -> setMessage ""
	rcontinue
    (KeyChar '~') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filter: " "~"
	case v of 
	    Just f -> maybe (return ()) (\x -> modifyFilter (x:)) $ parseFilter f 
	    Nothing -> setMessage ""
	rcontinue
    (KeyChar 's') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filename: " ""
	setMessage "feature coming soon!"
	rcontinue
    (KeyChar 'S') -> do
	(ys,xs) <- scrSize
	v <- commandRead (getKey $ return ()) stdScr (ys - 1) "Filename: " ""
	setMessage "feature coming soon!"
	rcontinue
    (KeyChar 'f') -> do
	p <- sel_puff
	case p of
	    Nothing -> setMessage "No puff selected" 
	    Just p -> do
		pkw <- fmap (concat . (map words)) $ varLookupAll "PRESERVED_KEYWORDS" 
		let kws = filter (`elem` pkw) (getFragmentStrings p "message.keyword")
		composePuff (cats p) kws
	continue
    (KeyChar 'p') -> composePuff [] [] >> rcontinue
    (KeyChar 'r') -> do
	p <- sel_puff
	case p of
	    Nothing -> return ()
	    Just p -> do
		pkw <- fmap (concat . (map words)) $ varLookupAll "PRESERVED_KEYWORDS" 
		let kws = filter (`elem` pkw) (getFragmentStrings p "message.keyword")
		composePuff [catParseNew (getAuthor p)] kws 
	continue
    (KeyF 1) -> (withLock screenLock helpScreen) >> hcontinue
    (KeyChar '\ESC') -> (withLock screenLock helpScreen) >> hcontinue
    (KeyChar n) | n `elem` "123456789" -> gfs (ord n - ord '1') >> rcontinue
    key -> keyError "Invalid key" key >> rcontinue
    
    where
	keyError s k = let ks = case k of (KeyChar key) -> show key; key-> show key in 
	    setMessage (s ++ ": " ++ ks ++ " -- " ++ helpText) 
	setMessage m = writeIORef message_r m
	getKey action = readChan ic >>= \v -> case v of
	    (Right p) -> do
		s <- varLookupElse "ON_INCOMING_PUFF" (return "")
		insertKeys ic s 
		addPuff p >> action >> getKey action
	    (Left k) -> return k
	continue = mainLoop gc ic yor psr selected_r filter_r next_r message_r cache_render_r cache_filter_r mark_set_r redraw_r
	rcontinue = continue
	hcontinue = (getKey $ return ()) >> rcontinue
	redraw = withLock screenLock $ (erase >> doRedraw >> refresh)
	scrollPuffs x = modifyIORef yor (+ x) >> clip_yo 
	gfs n = do
	    mark_set <- readIORef mark_set_r
	    modifyFilter (const (mark_set ! n))
	filter_thread = do
	    p <- sel_puff
	    case p of 
		Just (Puff {cats = [c]}) -> modifyFilter (FilterCategory c:)
		Just (Puff {cats = cs}) -> modifyFilter (FilterOr (map FilterCategory cs):)
		_ -> return ()
	sel_puff = do
	    ps <- readMVar psr
	    selected <- readIORef selected_r
	    return $ lookup selected ps 
	addPuff p = do
	    v <- varLookupAll "BEEP"
	    let f = FilterOr (catMaybes (map parseFilter v))
	    --putLogLn ("BEEP filter" ++ showFilter f)
	    when (filterAp f p) Curses.beep
	    n <- readIORef next_r
	    modifyIORef next_r (+1)
	    modifyMVar_ psr (\ps -> return ((n,p):ps)) 
	    writeIORef cache_render_r ([], -1) 
	    select_perhaps select_next
	getRendered = do
	    (ys,xs) <- scrSize
	    (cr,xs') <- readIORef cache_render_r
	    if xs == xs' then return cr else do
		ps <- readMVar psr 
		fs <- readIORef filter_r
		let f p = all (`filterAp` p) fs
		let nps =  [(x,rPuff xs p)| x@(_,p) <- ps] 
		writeIORef cache_render_r ((nps, xs)) 
		writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
		return nps
	showDestination cs kwds = (map catShowNew cs ++ map ('/':) kwds)
	readDestination :: String -> ([Category],[String])
	readDestination s = splitEither (map pi (simpleUnquote s)) where
	    pi ('/':s) = Right s
	    pi x = Left (catParseNew x)

	composePuff :: [Category] -> [String] -> IO ()
	composePuff cs kwds = do
	    e <- getEditor
	    fn <- getTmpFile
	    let it = ["To: " ++ simpleQuote (showDestination cs kwds) , "--------", ""] 
	    writeRawFile fn (stringToBytes $ unlines it)
	    withProgram $ System.system (e ++ " +3 " ++ fn)
	    pn <- fmap (lines . bytesToString)$ readRawFile fn 
	    if (length pn > 1 && pn /= it) then do 
		let (cs',kwds') = readDestination (drop 4 (head pn))
		let p = emptyPuff { cats = cs', fragments = [("message/body",FragmentText (unlines $ (drop 2 pn)))] ++ [ ("message.keyword",FragmentText k) |k <- kwds']}
		--erase >> putStrCenter stdScr "Sending Puff..." >> refresh
		sendPuff p else 
		    setMessage "Puff aborted."
	sendPuff p = do
	    p' <- expandAliases (cats p)
	    p <- return $ p {cats = p'}
	    gn <- varLookupAll "GALE_NAME"
	    t <- getClockTime
	    n <- case gn of 
		(n:_) -> return [("message/sender",FragmentText n)]
		_ -> return []
	    gi <- getGaleId
	    let sig = [Signed (Key gi [])]
	    idText <- getIdInstance
	    let ef = [("id/class",FragmentText fullName),("id/time",FragmentTime t),("id/instance", FragmentText idText)] ++ n
	    forkIO $ galeSendPuff gc (p {fragments = fragments p ++ ef, signature = sig})
	    setMessage "Puff being sent."
	    return ()
	modifyFilter f = do
	    modifyIORef filter_r f 
	    fs <- readIORef filter_r
	    setMessage (unwords (reverse $ map showFilter fs))
	    nps <- getRendered
	    let f p = all (`filterAp` p) fs
	    writeIORef cache_filter_r [x| x@((_,p),_) <- nps, f p]
	    select_perhaps select_next >> select_perhaps select_prev
	    center_puff
	    clip_yo
	getFiltered = do
	    getRendered
	    readIORef cache_filter_r
	buf_size = getFiltered >>= \ps -> return $ sum [x| (_,(x,_)) <- ps]
	center_puff = do
	    (ys,xs) <- scrSize
	    selected <- readIORef selected_r
	    ps <- liftM reverse getFiltered
	    let f t (((n,_),(h,_)):_) | n == selected = (t,t+h)
		f t ((_,(h,_)):rest) = f (t + h) rest
		f _ [] = (1000000,0)
	    let (mn,mx) = (f 0) ps
	    modifyIORef yor (min mn)
	    modifyIORef yor (max (mx - (ys - 1) ))
	scroll_end = do
	    (ys,xs) <- scrSize
	    bs <- buf_size
	    writeIORef yor (max (bs - (ys - 1)) 0)
	clip_yo = do
	    (ys,xs) <- scrSize
	    bs <- buf_size
	    modifyIORef yor (max 0)
	    modifyIORef yor (min (max 0 (bs - (ys - 1))))
	doRedraw = do 
	    (ys,xs) <- scrSize
	    selected <- readIORef selected_r
	    ps <- getFiltered
	    yo <- readIORef yor
	    m <- readIORef message_r
	    let rp _ [] =  return () 
		rp y (((i,_),(n,f)):rest) = do
		    when ((0,ys) `ovl` (y - yo, y - yo + n)) $ f stdScr (y - yo) (i == selected)
		    rp (y + n) rest
	    attempt (rp 0 (reverse ps))
	    -- draw status bar
	    wmove stdScr (ys - 1) 0
	    wClrToEol stdScr
	    standout >> waddstr stdScr (' ':m) >> standend
	    let s = show selected ++ "/" ++ show (length ps)
	    wmove stdScr (ys - 1) (xs - length s - 2)
	    wAttrOn stdScr attrBold >> waddstr stdScr s >> wAttrOff stdScr attrBold
	select_perhaps a = do
	    ps <- getFiltered
	    n <- readIORef selected_r
	    unless (maybe False (const True) $ lookup n (fsts ps)) a
	select_prev = do
	    ps <- liftM fsts getFiltered
	    let f sel ((n,p):_) | n < sel = n 
		f sel (_:ps) = f sel ps
	        f sel [] = sel
	    modifyIORef selected_r (\s -> f s ps)
	select_next = do
	    ps <- liftM fsts getFiltered
	    let f sel _ ((n,p):ps) | n > sel = f sel n ps
		f sel bg ((n,_):ps) | n > sel = f sel bg ps
	        f _ bg _ = bg
	    modifyIORef selected_r (\s -> f s s ps)



getAuthor p = maybe "unknown" id (getSigner p)

getSigner p = ss (signature p) where
	    ss (Signed (Key n _):_) = Just n
	    ss (Unverifyable (Key n _):_) = Just n
	    ss (_:sig) = ss sig
	    ss [] = Nothing

rPuff :: Int -> Puff -> (Int, Window -> Int -> Bool -> IO ())
rPuff mw p@(Puff {cats =cats, fragments = frags}) = (n + 2,f) where
    reformat s | length s < mw = [s]
    reformat s = let (a,b) = splitAt mw s in a : reformat b
    sender = case getFragmentString p "message/sender" of 
	Just s -> s ++ " <" ++ getAuthor p ++ ">" 
	Nothing -> "<" ++ getAuthor p ++ ">"
    body = case lookup "message/body" frags of
	Just (FragmentText s) -> lines s
	_ -> []
    time = case lookup "id/time" frags of 
	Just (FragmentTime t) -> Just t
	_ -> Nothing
    kwds = getFragmentStrings p "message.keyword"
    body' = concatMap reformat body
    n = length body'
    addCat w (x,y) = wAttrOn w attrBold >> waddstr w x >> wAttrOff w attrBold >> waddstr w ('@':y) 
    f w y selected = do
	when (y >= 0) $ do
	    wmove w y 0
	    sequence (intersperse (space w) $ map (addCat w) cats)
	    space w >> space w
	    waddstr w (unwords (map ('/':) kwds))
	    st <- case time of 
		Just t -> liftM ft $ toCalendarTime t
		Nothing -> return ""
	    let lst = length st
	    wAttrOn w attrBold
	    mvwaddstr w y (mw - length sender - lst - 1) sender
	    wAttrOff w attrBold
	    when selected $ void standout
	    mvwaddstr w y (mw - lst) st
	    when selected $ void standend
	mapM_ (\(p,s) ->  mvwaddstr stdScr (y + p) 0 s) (zip [1..] $ body')
	wAttrOn w attrBold
	mvwaddstr w (y + 1 + n) 0 (replicate mw $ if selected then '=' else '-')
	wAttrOff w attrBold
	return ()
    ft (CalendarTime {ctHour = h, ctMin = m, ctSec = s}) = fn h ++ ":" ++ fn m ++ ":" ++ fn s
    fn n = let x = show n ; lx = length x in if lx < 2 then replicate (2 - lx) '0' ++ x else x

	
(a,b) `ovl` (x,y) | y < a = False
(a,b) `ovl` (x,y) | b < x = False
_ `ovl` _ = True


------------------
-- Curses routines
------------------

addstr s = Control.Exception.try (addStr s) >> return ()
waddstr w s = Control.Exception.try (wAddStr w s) >> return ()
mvwaddstr w y x s = Control.Exception.try (mvWAddStr w y x s) >> return ()
wmove w y x = Control.Exception.try (wMove w y x) >> return ()
cursset x = Control.Exception.try (cursSet x) >> return ()
addln = Control.Exception.try addLn >> return ()

space w = waddstr w " "

--------------
-- help screen
--------------

buildTable :: [(String,String)] -> [String]
buildTable ps = map f ps where
    f (x,y) = replicate (bs - length x) ' ' ++ x ++ replicate 4 ' ' ++ y
    bs = 2 + maximum (map (length . fst) ps)



helpTable = [
    ("Keys:", ""),
    ("j","next puff"),
    ("k","previous puff"),
    ("G","last puff"),
    ("g","first puff"),
    ("<enter> ^E","forward one line"),
    ("<backspace> ^Y","back one line"),
    ("^F <PgDown>","forward one page"),
    ("^B <PgUp>","back one page"),
    ("^D","forward one half-page"),
    ("^U","back one half-page"),
    ("",""),
    ("c ~","add new filter"),
    ("u","pop one filter"),
    ("U","pop all filters"),
    ("!","invert filter at top of stack"),
    ("x","swap top two filters on stack"),
    ("a","filter to current author"),
    ("t","filter to current thread"),
    ("T","filter to strict thread"),
    ("m[1-9]","set mark - save current filter stack at given key"),
    ("[1-9]","recall filter stack saved here with 'm'"),
    ("",""),
    ("f","follow-up to category"),
    ("r","reply to author"),
    ("p","compose new puff"),
    ("",""),
    ("s","save selection to file"),
    ("S","save current puffs to file"),
    ("^R","reconnect to all servers"),
    ("E","edit and reload configuration file"),
    ("q","quit program"),
    ("^L","redraw screen"),
    ("",""),
    ("Filters:",""),
    ("~a","author"),
    ("~s","puffers real name"),
    ("~c","category"),
    ("(default)","substring"),
    ("~k","keyword"),
    ("~t","thread - transitive closure of categories")
    ]

helpScreen = do
    (ys,xs) <- scrSize
    let s = "Help Screen - press any key to exit."
    erase
    mvwaddstr stdScr 0 ((xs - length s) `div` 2 ) s
    move 2 0
    mapM_ (\s -> addstr s >> addln) (buildTable helpTable)

    refresh
    
helpText = "Need Help? Press F1"

-------------------------
-- options/initialization
-------------------------

env = Env { 
    envAction = return (),
    envVerbose = False, 
    envConfig = Nothing,
    envJustArgs = False,
    envNoPufflog = False,
    envErrorLog = Nothing
    } 


options :: [OptDescr (Env -> Env)]
options = [
	Option ['v']  ["verbose"] (NoArg (\e->e{envVerbose = True})) "verbose output to errorlog",
	Option ['V']  ["version"] (NoArg (doAction doVersion)) "print version information",
--	Option ['c']  ["config"] (ReqArg (\x e->e{envConfig = (Just x)}) "FILE") "use this configuration file",
	Option ['s']  ["sample"] (NoArg (doAction doSample)) "print sample configuration file to stdout",
	Option ['m']  ["man"] (NoArg (doAction doMan)) "print all internal help screens to stdout",
	Option ['e']  [] (NoArg (\e->e{envJustArgs = True})) "only subscribe to command line arguments",
	Option ['P']  [] (NoArg (\e->e{envNoPufflog = True})) "do not read or write pufflog",
	Option []  ["errorlog"] (ReqArg (\x e->e{envErrorLog = Just x}) "FILE") "log errors to file",
	Option []  ["testCurses"] (NoArg (doAction (cursesTest >> exitSuccess))) "print curses diagnostics",
	Option []  ["dumpKey"] (ReqArg (\x -> doAction (dumpKey x >> exitSuccess)) "KEYFILE") "print info for keyfile"
    ]

void a = a >> return ()

ginsuOpts :: IO (Env,[String])
ginsuOpts = do
    args <- getArgs
    r@(env,_) <- case (getOpt Permute options args) of
	   (as,n,[]) -> return (foldr ($) env as ,n)
    	   (_,_,errs) -> putErrDie (concat errs ++ usage)
    envAction env
    handle (\_ -> putErrDie ("Could not find configuration file: ~/.gale/ginsu.config\n" ++ trailer)) (loadStdConfiguration ".gale/ginsu.config")
    return r

doVersion = putStrLn fullName >> exitSuccess
doSample = putStr exampleConf >> exitSuccess
doMan = putStrLn usage >> putStrLn (unlines (buildTable helpTable)) >> exitSuccess

usage = usageInfo header options ++ trailer where
    header = "Usage: ginsu [OPTION...] categories..."
trailer = "\nTo get started, generate a configuration file using the following command:\nmkdir ~/.gale ; ginsu -s > ~/.gale/ginsu.config\nEdit the configuration file and run ginsu!\n\n"
 ++ "The homepage is at <http://repetae.net/john/computer/ginsu/>\nTo report a bug go to <http://bugs.ofb.net/cgi-bin/bugzilla/enter_bug.cgi?product=Ginsu>\nWhen reporting a bug, include the last few lines of ~/.gale/ginsu.errorlog if it seems appropriate."

doAction :: IO a -> (Env -> Env)
doAction a = \e -> e {envAction = a >> envAction e}



commandRead :: IO Curses.Key -> Window -> Int -> String -> String -> IO (Maybe String)
commandRead ic win yloc prompt init = do
    cursset Visible
    v <- cr (reverse init) 
    cursset Invisible
    return v
    where
    cr v = pc v >> ic >>= \x -> case x of
	KeyEnter -> return $ Just (reverse v)
	(KeyChar '\n') -> return $ Just (reverse v)
	(KeyChar '\r') ->  return $ Just (reverse v)
	KeyBackspace -> cr (drop 1 v)
	(KeyChar c) -> cr (c:v)
	_ -> return Nothing
    pc v = do
	wmove win yloc 0
	wClrToEol win
	wAttrOn win attrBold
	waddstr win prompt 
	wAttrOff win attrBold
	waddstr win (reverse v)
	wmove win yloc (length v + length prompt)
	refresh
